#include "mux.h"

/* make a mux_stream using a dummy thread/tstream */
/* called when we get a SEXY_NEW packet */
thread mux_stream_new(mux_t mux, int fd, xmlstream_onNode evt)
{
    thread t;
    tstream ts;

    log_debug(ZONE,"mux_stream_new %d",fd);

    /* create a new thread for this connection */
    t = thread_new();
    ts = pmalloc(t->p, sizeof(_tstream));
    /* set the init values */
    memset(ts, '\0',sizeof(_tstream));
    ts->data = (void *)evt;
    ts->fd = fd;
    ts->data5 = (void *)mux;
    t->data = (void *)ts;

    mux_stream(t, NULL, 0); /* init */

    return t;
}

/* write an xmlnode to the polld client */
int mux_stream_write(thread t, xmlnode x)
{
    char *data;

    log_debug(ZONE,"mux_stream_write");

    if(t == NULL || x == NULL)
        return 0;

    /* convert the xmlnode to a string and sent it to the polld client */
    data = xmlnode2str(x);
    mux_write(SEXY_WRITE,t,data,strlen(data));

    return 0;
}

/* give a standard template xmlnode to work from */
xmlnode mux_stream_header(char *namespace, char *to, char *from)
{
    xmlnode x;
    char id[10];

    sprintf(id,"%X",(int)time(NULL));

    x = xmlnode_new_tag("stream:stream");
    pool_label(xmlnode_pool(x),"XShead",0);
    xmlnode_put_attrib(x, "xmlns:stream", "http://etherx.jabber.org/streams");
    xmlnode_put_attrib(x, "id", id);
    if(namespace != NULL)
        xmlnode_put_attrib(x, "xmlns", namespace);
    if(to != NULL)
        xmlnode_put_attrib(x, "to", to);
    if(from != NULL)
        xmlnode_put_attrib(x, "from", from);

    return x;
}

/* special wrapper to only write the root opening */
void mux_stream_write_header(thread t, xmlnode x)
{
    spool s;
    char *fixr, *head;

    if(xmlnode_has_children(x))
        return;

    s = spool_new(x->p);
    spooler(s,"<?xml version='1.0'?>",xmlnode2str(x),s);
    head = spool_print(s);
    fixr = strstr(head,"/>");
    *fixr = '>';
    ++fixr;
    *fixr = '\0';
    log_debug(ZONE,"head: %s",head);

    mux_write(SEXY_WRITE,t, head, strlen(head));
    xmlnode_free(x);
}

/* special wrapper to send an error and close this stream */
void mux_stream_close(thread t)
{
    mux_write(SEXY_WRITE, t, "</stream:stream>", 16);
    mux_close(t);
}

/* special wrapper to send an error and close this stream */
void mux_stream_error(thread t, char *err)
{
    mux_write(SEXY_WRITE,t, "<stream:error>", 14);
    mux_write(SEXY_WRITE,t, err, strlen(err));
    mux_write(SEXY_WRITE,t, "</stream:error>",15);
    mux_stream_close(t);
}

void _mux_stream_startElement(void* userdata, const char* name, const char** atts)
{
    thread t = (thread)userdata;
    tstream ts = (tstream)(t->data);
    xmlnode current = (xmlnode)(ts->data2);
    xmlstream_onNode onNode = (xmlstream_onNode)(ts->data);
    pool p;

    if(current == NULL)
    {
        p = pool_heap(5*1024); /* 5k, typically 1-2k each plus copy of self and workspace */
        current = xmlnode_new_tag_pool(p,name);
        xmlnode_put_expat_attribs(current, atts);

        if(ts->flag == 0)
        { /* this is a special flag saying that this is the <stream> root */
            ts->flag = 1;
            ts->data2 = current; /* so that the cleanup can free it if we exit in the handler */
            (onNode)(t, current, XMLSTREAM_ROOT);
            xmlnode_free(current);
            current = NULL;
            pool_label(xmlnode_pool(current),"XSroot",0);
        }else{
            pool_label(xmlnode_pool(current),"XSnode",0);
        }
    }
    else
    {
        current = xmlnode_insert_tag(current, name);
        xmlnode_put_expat_attribs(current, atts);
    }

    ts->flag++;
    if(ts->flag > XML_MAX_DEPTH)
        mux_stream_error(t, "Maximum XML Element Depth Reached");

    ts->data2 = current;
}


void _mux_stream_endElement(void* userdata, const char* name)
{
    thread t = (thread)userdata;
    tstream ts = (tstream)(t->data);
    xmlnode current = (xmlnode)(ts->data2);
    xmlstream_onNode onNode = (xmlstream_onNode)(ts->data);
    xmlnode parent;

    /* if it's already NULL we've received </stream>, tell the app and we're outta here */
    if(current == NULL)
    {
        (onNode)(t, current, XMLSTREAM_CLOSE);
        mux_close(t);
        return;
    }else{
        parent = xmlnode_get_parent(current);

        /* we are the top-most node, feed to the app who is responsible to delete it */
        if(parent == NULL)
            (onNode)(t, current, XMLSTREAM_NODE);

        ts->data2 = parent;
    }
    ts->flag--;
}


void _mux_stream_charData(void* userdata, const char* s, int len)
{
    thread t = (thread)userdata;
    tstream ts = (tstream)(t->data);
    xmlnode current = (xmlnode)(ts->data2);

    if(current == NULL)
    {
        /* we must be in the root of the stream where CDATA is irrelevant */
        return;
    }

    xmlnode_insert_cdata(current, s, len);
}

void _mux_stream_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    XML_Parser p = (XML_Parser)(ts->data1);
    xmlnode current = (xmlnode)(ts->data2);

    log_debug(ZONE,"mux_stream_cleanup");
    if(current != NULL)
        xmlnode_free(current); /* cleanup anything left over */
    XML_ParserFree(p);
}

/* read data from a polld connection */
void mux_stream(thread t, void *buffer, int bufsz)
{
    tstream ts = (tstream)(t->data);
    xmlstream_onNode onNode = (xmlstream_onNode)(ts->data);
    XML_Parser p = (XML_Parser)(ts->data1);
    xmlnode n;

    if(buffer == NULL)
    { /* init */
        log_debug(ZONE,"mux_stream_main: startup");
        pool_cleanup(t->p, _mux_stream_cleanup, (void *)t);
        p = XML_ParserCreate(NULL);
        ts->data1 = (void *)p;
        ts->data2 = NULL;
        XML_SetUserData(p, (void *)t);
        XML_SetElementHandler(p, (void *)_mux_stream_startElement, (void *)_mux_stream_endElement);
        XML_SetCharacterDataHandler(p, (void *)_mux_stream_charData);
/*        (onNode)(t, NULL, XMLSTREAM_INIT); */
        return;
    }

    log_debug(ZONE,"mux_stream_main: processing");

    if(!XML_Parse(p, buffer, bufsz, 0))
    {
        log_warn("etherx","XML Stream parsing error '%s' from host '%s'",(char *)XML_ErrorString(XML_GetErrorCode(p)),ts->host);
        (onNode)(t, NULL, XMLSTREAM_ERROR);
        mux_close(t);
        return;
    }

    n = (xmlnode)(ts->data2);
    if(pool_size(xmlnode_pool(n)) > 1000000)
    {
        log_warn("etherx","XML Stream node size overrun from fd %d",ts->fd);
        (onNode)(t, NULL, XMLSTREAM_ERROR);
        mux_close(t);
        return;
    }
}

