#include "polld.h"

int conncnt=0,conns=0;
int maxi;
struct pollfd client[MY_OPEN_MAX];
int needfork=0;

/* create a new socket to connect to svc_mux */
int make_masterfd(char *host, int port)
{
    int sockfd;
    int flags;
    struct sockaddr_in servaddr;

    if(host == NULL)
    {
        printf("No Host...\n");
        return -1;
    }

    if( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) <0)
    {
        printf("Socket Create Error!\n");
        return -1;
    }

    /* create the addr struct, and clear the mem */
    memset(&servaddr, '\0', sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(port);
    inet_pton(AF_INET, host, &servaddr.sin_addr);

    flags = fcntl(sockfd, F_GETFL, 0);
    flags |= O_NONBLOCK;
    if(fcntl(sockfd, F_SETFL, flags) <0)
    	printf("Error setting master flags\n");

    /* connect to svc_mux */
    connect(sockfd, (SA *)&servaddr, sizeof(servaddr));

    if(errno == ECONNREFUSED)
        return -1;
    if(sockfd >0)
        return sockfd;

    return -1;
}

/* main polld func */
int main(int argc, char *argv[])
{
    int i, listenfd, connfd, sockfd;
    int nready;
    int flags;
    int opt=TRUE;
    int id;
    size_t n;
    char line[1024];
    socklen_t clilen;
    struct sockaddr_in cliaddr, servaddr;
    int masterfd;

    /* connect to svc_client on port 5739 */
    masterfd = make_masterfd("127.0.0.1", 5739);
    if(masterfd < 0)
    {
        printf("Error creating master socket.  Exitting!\n");
        exit(1);
    }
    printf("MasterFD: %d\n", masterfd);
    /* create a listening socket for port 5222 */
    listenfd = socket(AF_INET, SOCK_STREAM,0);

    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(SERV_PORT);

    if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, sizeof(opt))<0)
        printf("Error setsockopt\n");


    /* bind the socket to the local part, and listen for connections */
    bind(listenfd, (SA *)&servaddr, sizeof(servaddr));
    listen(listenfd, LISTENQ);

    flags = fcntl(listenfd, F_GETFL, 0);
    flags |= O_NONBLOCK;
    if(fcntl(listenfd, F_SETFL, flags) <0)
        printf("Error setting flags\n");

    /* set the initial client[] values for the listenfd, and masterfd */
    client[0].fd = listenfd;
    client[0].events = POLLRDNORM;
    client[1].fd = masterfd;
    client[1].events = POLLRDNORM;
    for(i = 2; i<MY_OPEN_MAX; i++)
        client[i].fd = -1;

    /* create a new stream from the fd and link it to sessions */
    stream_new(masterfd);

    /* last fd we use */
    maxi = 1;

    /* this polld id */
    id = 1;
    while(1)
    {
        /* if we have over 1K conns */
        if(needfork)
        {
            needfork = 0;
            if(fork()==0)
            {
                printf("Forked\n");
                /* new polld id */
                id = id + 1;
                /* reset values */
                conns = 0;
                needfork = 0;
                maxi = 3;
                /* create a new connection to svc_client */
                masterfd = make_masterfd("127.0.0.1", 5739);
                if(masterfd < 0)
                {
                    printf("Error creating master socket.  Exitting!\n");
                    exit(1);
                }
                printf("MasterFD: %d\n", masterfd);
                client[1].fd = masterfd;
                client[1].events = POLLRDNORM;
                for(i = 2; i<MY_OPEN_MAX; i++)
                    client[i].fd = -1;
                /* set sessions to NULL, create a new stream from the masterfd */
                stream_reset();
                stream_new(client[1].fd);
            }
            else
            {
                /* the old polld no longer listens on 5222 */
                client[0].fd = -1;
            }
        }
        printf("[%d]Conn Count:  %d/%d/%d\n", id, conns, conncnt, maxi);
        /* check if we need to do something, wait until we do */
        nready = poll(client, maxi + 1, INFTIM);

        /* if we have a new socket to accept */
        if(client[0].revents & POLLRDNORM) {
            clilen = sizeof(cliaddr);
            /* accept the socket, and associate a new mux with it */
            connfd = accept(listenfd, (SA *)&cliaddr, &clilen);
            flags = fcntl(connfd,F_GETFL, 0);
            flags |= O_NONBLOCK;
            if(fcntl(connfd, F_SETFL, flags) <0)
                printf("Error setting flags\n");
            printf("[%d]Conn...master %d...", id, client[1].fd);
            /* tell mux about the new connection */
            polld_mux_new(client[1].fd, connfd);
            
            /* keep going if there are more to accept */
            if(--nready <=0)
                continue;
        }
        if(client[1].revents & POLLRDNORM) {
            /* muxxy now */
            /*
            
            int data_sz;
            char buf[1024];

            data_sz = read(client[1].fd, &buf, 1024);
            if(data_sz <= 0 || errno == ECONNRESET)
            {
                printf("Master server connection has failed...exitting\n");
                exit(1);
            }
            printf("Got:  %s\n", buf); 

            */
            //printf("Data for MasterFD...");
            polld_mux_read(client[1].fd);
            
        } else if(client[1].revents & POLLOUT) {
            stream s;

            //printf("Master Write...");
            s = stream_find_by_fd(client[1].fd);
            stream_write(s, NULL, 0);
            if(s->write == NULL)
                client[1].events ^= POLLOUT;
        } else if( client[1].revents & POLLERR) {
            if(errno == EWOULDBLOCK)
                break;
            printf("We got an error on the masterfd... not good\n");
            printf("Error: %s\n", strerror(errno));
        }
        else if(client[1].revents & (POLLNVAL | POLLHUP))
        {
            printf("Master server connection has failed...exitting\n");
            exit(1);
        }

        for(i = 2;i<=maxi;i++) {
            /*printf("Checking %d...",i);*/
            if((sockfd = client[i].fd)<0 || client[i].revents == 0)
            {
                continue;
            }
            if(client[i].revents & (POLLRDNORM | POLLERR)) {
                stream s;
                //printf("Searching for %d...", client[i].fd);
                s = stream_find_by_fd(client[i].fd);
                if(s==NULL)
                {
                    printf("ERROR! No Stream\n");
                    break;
                }
                if(s->xmlstream == NULL)
                {
                    printf("XS Not Set\n");
                    client[i].revents = 0;
                    continue;
                }                
                if((n = read(sockfd, &line, MAXLINE)) < 0) {
                    if (errno == ECONNRESET) {
                        printf("Conn Reset!\n");
                        polld_mux_error(client[1].fd, stream_find_by_fd(sockfd));
                        conncnt--;
                        conns--;
                        close(sockfd);
                        stream_unlink(sockfd);
                        client[i].fd = -1;
                    } else
                        printf("readling error\n");
                } else if (n == 0) {
                    printf("No Data, Gone\n");
                    polld_mux_error(client[1].fd, stream_find_by_fd(sockfd));
                    close(sockfd);
                    conncnt--;
                    conns--;
                        stream_unlink(sockfd);
                    client[i].fd = -1;
                } else {
                    /* Using mux now
                     * writen(masterfd, line, n);
                     */
                    polld_mux_write(client[1].fd, s, (char *)&line, n);
                }

                if(--nready <= 0)
                    break;
            }
            else if(client[i].revents & POLLOUT)
            {
                stream s;

                //printf("Need to Write...");
                s = stream_find_by_fd(client[i].fd);
                if(s == NULL)
                    continue;
                stream_write(s, NULL, 0);
                if(s->write == NULL)
                    client[i].events ^= POLLWRNORM;
            }
            else if(client[i].revents & (POLLNVAL | POLLHUP))
            {
                conncnt--;
                conns--;
                close(sockfd);
                stream_unlink(sockfd);
                client[i].fd = -1;
                printf("Cleaned up socket\n");
            }
            else
            {
                printf("Unhandled Event: %d\n", client[i].revents);
            }
        }
    }

    return 0;
}

/* new accepted threads get sent here */
void polld_mux_new(int mfd, int newfd)
{
    stream s;
    sexy_t packet;

    printf("New for %d passed %d\n",client[1].fd, mfd);

    /* create a new sexy_t packet, to sent to svc_mux */
    packet = malloc(sizeof(_sexy_t));
    packet->type = SEXY_NEW;
    packet->size = 0;
    packet->fd = newfd;
    packet->data = NULL;

    /* find the masterfd (connection to svc_mux) */
    s = stream_find_by_fd(mfd);
    if(s == NULL)
    {
        printf("Cannont make new, error finding stream\n");
        return;
    }
    /* send it the SEXY_NEW packet, telling it about the new connection */
    stream_write(s, packet, sizeof(_sexy_t));

    free(packet);

    return;
}

void polld_mux_write(int fd, stream s, char *data, int sz)
{
    sexy_t packet;

    if(s->xmlstream == NULL)
    {
        printf("Invalid XS\n");
        return;
    }

    packet = malloc(sizeof(_sexy_t));
    packet->type = SEXY_READ;
    packet->size = sz;
    packet->fd = s->fd;
    packet->data = s->xmlstream;

    stream_write(stream_find_by_fd(client[1].fd), packet, sizeof(_sexy_t));
    stream_write(stream_find_by_fd(client[1].fd), data, packet->size);

    client[1].events |= POLLOUT;

    free(packet);

    return;
}

/* read data off the masterfd */
void polld_mux_read(int fd)
{
    char data[1024];
    size_t szread;
    stream s;

    szread = read(fd, &data, 1023);
    if(errno == ECONNRESET)
    {
        printf("Master server connection has failed...exitting\n");
        exit(1);
    }
    
    /* find the masterfd session */
    s = stream_find_by_fd(fd);
    if(s == NULL)
    {
        printf("No session for read!\n");
        return;
    }
    stream_master_read(fd, s, &data, szread);
    /*if(packet->size > 0)
    {
        data = malloc(packet->size);
        szread = readn(fd, data, packet->size);
        if(szread != packet->size)
        {
            printf("Error read mux data!  Got:  %s\n", data);
            return;
        }
    }

    polld_mux_parse(fd, packet, data);

        free(data); 
    free(packet); */

    return;
}

/* parse a packet from svc_mux */
void polld_mux_parse(int fd, sexy_t packet, char *data)
{
    stream s;
    int i;

    /* what sexy_t type did we get?? */
    switch(packet->type)
    {
        /* sent to polld when mux accepted the new client connection? */
        case SEXY_SET:
            for(i=2;i<MY_OPEN_MAX;i++)
                /* find an open slot, and put it in here */
                if(client[i].fd <0)
                {
                    stream new;

                    printf("Accepting new connection...");
                    printf("From %d...", client[1].fd);
                    new = stream_new(packet->fd);
                    /* set the fd of this slot */
                    client[i].fd = packet->fd;
                    /* inc our connections count */
                    conncnt++;
                    conns++;
                    /* if we have too many, flag that we need a fork */
                    if(conncnt > 0 && conncnt % 1000 == 0)
                        needfork = 1;
                    /* keep the xmlstream pointer in this stream */
                    new->xmlstream = packet->data;
                    printf("[%d]...done\n", i);
                    break;
                }
            /* if somehow we reached 1024 open connections on this polld */
            if(i == MY_OPEN_MAX)
                exit(1);

            /* mark that we want to be notified of read events */
            client[i].events = POLLRDNORM;
            if(i > maxi)
                maxi = i; 

            break;
        /* we want to write data from svc_mux to the client stream */
        case SEXY_WRITE:
            /* get the right stream */
            s = stream_find_by_fd(packet->fd);
            if(s == NULL)
                break;
            /* write the packet to the stream */
            stream_write(s, data, packet->size);
            /* if there is still data to write */
            if(s->write != NULL)
            {
                /* find the slot */
                for(i=1;i<MY_OPEN_MAX;i++)
                {
                    if(client[i].fd == s->fd)
                    {
                        /* mark that there is data to be written */
                        client[i].events |= POLLWRNORM;
                        break;
                    }
                }
            }
            break;
        /* this session has been closed */
        case SEXY_CLOSE:
            /* get the stream */
            s = stream_find_by_fd(packet->fd);
            if(s == NULL)
                break;
            /* printf an error message */
            polld_mux_error(fd, s);
            /* kill the connection and unlink this stream */
            close(s->fd);
            stream_unlink(fd);
            break;
        default:
            printf("Unhandled type = %d\n", packet->type);
            break;
    }

    return;
} 

/* there was an error on the stream */
void polld_mux_error(int fd, stream s)
{
    stream master_stream;
    sexy_t packet;

    /* figure out which error it is */
    if(s->xmlstream == NULL)
    {
        printf("Invlid XS\n");
        return;
    }
    if(s->fd == 0)
    {
        printf("FD is Zero!\n");
        return;
    }
    if(s->fd == client[1].fd)
    {
        printf("Umm.... masterfd error!\n");
        return;
    }

    printf("Error on %d XS: %X\n",s->fd,s->xmlstream);

    packet = malloc(sizeof(_sexy_t));
    packet->type = SEXY_ERROR;
    packet->size = 0;
    packet->fd = s->fd;
    packet->data = s->xmlstream;

    master_stream = stream_find_by_fd(fd);
    stream_write(master_stream, packet, sizeof(_sexy_t));

    free(packet);

    return;
}
