#include <polld.h>

stream sessions=NULL;

/* write data to a stream */
void stream_write(stream cur, void *buffer, int bufsz)
{
    stream_buffer b;
    pool p;
    int data_sz;

    /* if there is data to write */
    if(buffer != NULL && bufsz != 0)
    {
        printf("New Write...");
        /*create a new stream buffer to write */
        p = pool_new();
        b = pmalloc(p, sizeof(_stream_buffer));

        if(bufsz < 0)
            bufsz = strlen(buffer);

        b->data = pmalloc(p, bufsz);
        memcpy(b->data, buffer, bufsz);
        b->data_sz = bufsz;
        b->sent = 0;
        b->next = NULL;
        b->p = p;

        /* add this stream buffer to the start of the stream buffer */
        if(cur->write == NULL)
            cur->write = b;
        else
        {
            b->next = cur->write;
            cur->write = b;
        }
    }

    /* write all the stream buffers to the stream fd */
    while(cur->write != NULL)
    {
        printf("Writing[%d]...", cur->fd);
        /* write all the data we can */
        data_sz = write(cur->fd, cur->write->data, cur->write->data_sz);
        if(data_sz == -1 && errno == EAGAIN)
        {
            printf("Error Writing\n");
            data_sz = 0;
        }
        if(data_sz < 0)
        {
            printf("Error writing\n");
            b = cur->write;
            cur->write=cur->write->next;
            pool_free(b->p);
            break;
        }
        /* if we wote all the data we needed to */
        if((data_sz + cur->write->sent) == cur->write->data_sz)
        {
            printf("Wrote %d bytes\n", data_sz); 
            b = cur->write;
            cur->write=cur->write->next;
            pool_free(b->p);
        /* otherwise, remember our position to continue the write */
        } else {
            cur->write->sent += data_sz;
            return;
        }
    }
}

/* read data off the masterfd (polld's connection to svc_mux) */
void stream_master_read(int masterfd, stream mux, void *data, int buffsz)
{
    void *buff;
    int newsz, leftsz;
    sexy_t sexy;

    /* reads a sexy_t packet off the socket, and adds it to the buffer */
    newsz = buffsz + (mux->buffsz - mux->cur);
    buff = malloc(newsz);
    memcpy(buff,mux->buff + mux->cur,mux->buffsz - mux->cur);
    memcpy(buff + (mux->buffsz - mux->cur),data,buffsz);

    /* ditch leftovers and reset */
    free(mux->buff);
    mux->buff = buff;
    mux->cur = 0;
    mux->buffsz = newsz;

    /* while there is still data in the buffer */
    while(mux->cur < mux->buffsz)
    {
        /* calcualte how much is left */
        leftsz = mux->buffsz - mux->cur;

        /* if there isn't enough to be a sexy_t packet */
        if(leftsz < sizeof(_sexy_t))
            break;

        /*otherwise turn the next chunk into a sexy_t packet */
        sexy = (sexy_t)(mux->buff + mux->cur);
        //printf("stream->cur %d/%d sexy->size %d...",mux->cur,mux->buffsz,sexy->size);
        /* if we didn't get the whole packet, don't try to read it */
        if(leftsz < sizeof(_sexy_t) + sexy->size)
            break;

        /* we've got a full pair, read and process */
        mux->cur += sizeof(_sexy_t);
        /* parse the packet, move the cur pointer to the right spot */
        polld_mux_parse(masterfd, sexy, mux->buff + mux->cur);
        mux->cur += sexy->size;
    }

    return;
}
    
/* create a new stream from a fd, and add it to the list */
stream stream_new(int fd)
{
    stream new;
    
    /* malloc the new stream, and link it to sessions */
    new = malloc(sizeof(_stream));
    new->fd = fd;
    new->write =NULL;
    new->next = sessions;
    new->xmlstream = NULL;

    sessions = new;

    return new;
}

/* unlink a stream from the sessions list */
void stream_unlink(int fd)
{
	stream next;
	stream prev;

	if(sessions == NULL || fd < 0)
        return;

	for(prev=NULL,next = sessions;next != NULL; next = next->next)
	{
		if(next->fd == fd)
		{
			if(prev ==NULL)
			{
				prev=next->next;
				sessions=prev;
			}
			else
				prev->next = next->next;
	    if(next != NULL)
			{
                /* XXX this is a mem leak 
                if(next->write)
                    pool_free(next->write->p);
                */
                while(next->write!=NULL)
                {
                    stream_buffer curr=next->write;
                    next->write=next->write->next;
                    pool_free(curr->p);
                }
				free(next);
                return;
			}
		}
		prev = next;
	}
}

/* find a stream, from its fd */
stream stream_find_by_fd(int fd)
{
    stream next;

    for(next = sessions; next != NULL; next=next->next)
    {
        if(next->fd == fd)
            return next;
    }
    return NULL;
}

/* when we fork, reset the sessions */
void stream_reset(void)
{
    sessions = NULL;
    return;
}
