#include "mux.h"

/*

chit chat time...

listen on 5765 for mux connections
set up a tstream on new muxs, with custom read, and setup pointer in the tstream
if they die (shouldn't ever happen) just flag that they died and let leak in case there were references

when reading from them, get three states:
    new connect:
        create a new thread struct & xmlstream, use custom svc_client
        new thread has pointer to it's mux
        pass back xmlstream pointer to polld as the data
    new data
        feed to xmlstream
        xmlstream feeds to custom svc_client
    dead connection
        normal cleanup for thread struct

custom svc_client is identical except for the write calls
write does xmlnode2str and a big write call over the mux connection
write checks for a valid mux

 */

/* flag this mux process as dead, leak it, might be referenced yet and this shouldn't happen anyway! */
void svc_mux_cleanup(void *arg)
{
    mux_t mux = (mux_t)arg;
    mux->t = NULL;
}

/* start a new mux thread */
void svc_mux_new(thread t, int sock, struct sockaddr_in sa)
{
    mux_t mux;

    /* malloc a new mux structure */
    mux = malloc(sizeof(_mux_t));
    mux->buffsz = mux->in = mux->out = mux->conns = mux->cur = 0;
    mux->buff = NULL;
    mux->t =  tstream_new(sock, inet_ntoa(sa.sin_addr), mux_read, NULL, (void *)mux);
    /* when the tstream dies, svc_mux_cleanup is called */
    pool_cleanup(mux->t->p,svc_mux_cleanup,mux);

    log_debug(ZONE,"new mux_t %X",mux);
}

/* main entry point for jserver */
void svc_mux(void)
{
    /* if there is a <mux/> flag in the config */
    if(js_config("mux") != NULL)
    {
        /* enabling mux means to use the 'custom' svc_client */
        log_debug(ZONE,"svc_mux ENABLED");
        /* listen on 5739, the custom svc_client port */
        if(tlisten_new(5739, "127.0.0.1", svc_mux_new, NULL) == NULL)
        {
            log_error("svc_mux","failed to listen on 127.0.0.1:5739");
        }else{
            log_debug(ZONE,"listening on 127.0.0.1:5739");
        }
    }
}
